from yta_general_utils.programming.validator.number import NumberValidator


class GraphicAxis:
    """
    Class that represent a Graphic axis with
    its min and max range.
    """

    range: tuple[float, float] = None
    """
    The range of the axis, a (min, max) tuple.
    """

    def __init__(self, min: float, max: float):
        if not NumberValidator.is_number(min) or not NumberValidator.is_number(max):
            raise Exception('The parameters "min" and "max" must be numbers.')
        
        if min >= max:
            raise Exception('The "min" parameter cannot be greater or equal than the "max" parameter.')

        self.range = (min, max)

    @property
    def min(self):
        """
        The minimum value.
        """
        return self.range[0]
    
    @property
    def max(self):
        """
        The maximum value.
        """
        return self.range[1]


import cv2


def detect_faces(image_filename):
    """
    This method receives the 'image_filename' image input and detects the faces that
    are in that image. It returns an array with all existing faces, with their coordinates
    and the width and height of the rectangle that surrounds the face.
    """
    # I used this: https://www.datacamp.com/tutorial/face-detection-python-opencv
    img = cv2.imread(image_filename)
    gray_image = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)

    face_classifier = cv2.CascadeClassifier(
        cv2.data.haarcascades + 'haarcascade_frontalface_default.xml'
    )

    face = face_classifier.detectMultiScale(
        gray_image, scaleFactor = 1.1, minNeighbors = 5, minSize = (40, 40)
    )

    faces_detected = []

    for (x, y, w, h) in face:
        faces_detected.append({
            'x': x,
            'y': y,
            'width': w,
            'height': h,
        })
        #cv2.rectangle(img, (x, y), (x + w, y + h), (0, 255, 0), 4)

    #cv2.imwrite('wip/screenshot_with_faces.png', img)
        
    return faces_detected
    
def detect_humans(image_filename):
    """
    This method receives the 'image_filename' image input and detects the humans that
    are in that image. It returns an array with all existing humans, with their coordinates
    and the width and height of the rectangle that surrounds the human.

    This works only for human full bodies.
    """
    # Reading the Image
    image = cv2.imread(image_filename)

    # initialize the HOG descriptor
    hog = cv2.HOGDescriptor()
    hog.setSVMDetector(cv2.HOGDescriptor_getDefaultPeopleDetector())

    # detect humans in input image
    (humans, _) = hog.detectMultiScale(image, winStride = (10, 10), padding = (32, 32), scale = 1.1)

    # getting no. of human detected
    print('Human Detected : ', len(humans))

    humans_detected = []

    # loop over all detected humans
    for (x, y, w, h) in humans:
        humans_detected.append({
            'x': x,
            'y': y,
            'width': w,
            'height': h,
        })
        pad_w, pad_h = int(0.15 * w), int(0.01 * h)
        cv2.rectangle(image, (x + pad_w, y + pad_h), (x + w - pad_w, y + h - pad_h), (0, 255, 0), 2)

    cv2.imwrite('wip/test_human_detection.png', image)

    return humans_detected


from yta_general_utils.programming.output import Output
from yta_general_utils.programming.validator import PythonValidator
from yta_general_utils.file.enums import FileTypeX, FileExtension
from yta_general_utils.downloader.audio import download_audio
from yta_general_utils.downloader.gif import download_gif
from yta_general_utils.downloader.google_drive import GoogleDriveResource
from yta_general_utils.downloader.image import download_image_2
from yta_general_utils.downloader.video import download_video
from yta_general_utils.downloader.utils import download_file
from yta_general_utils.dataclasses import FileReturn
from typing import Union


# TODO: Make all submethods return the element read (image as Pillow,
# video as VideoFileClip or numpy, etc. and also the filename
class Downloader:
    """
    Class to encapsulate the functionality related to download resources
    from the Internet.
    """

    # TODO: Maybe move the checkings to the specific 'download_x' method
    # and not here that is more a encapsulation class
    @staticmethod
    def download_audio(
        url: str,
        output_filename: Union[str, None] = None
    ):
        """
        Download the audio file from the provided 'url' and stores
        it locally as 'output_filename'.
        """
        if not PythonValidator.is_string(url):
            raise Exception('The provided "url" parameter is not a string.')

        # TODO: Add more checkings here (?)
        # I don't know the exception of the file
        return download_audio(url, Output.get_filename(output_filename, FileTypeX.AUDIO))
    
    @staticmethod
    def download_gif(
        query: str, 
        output_filename: Union[str, None] = None
    ):
        """
        Search for a gif with the provided 'query' and download it,
        if existing, to a local file called 'output_filename'.

        TODO: I think this is unexpected, because it is searching
        from Giphy and not downloading a file from a url as a gif...
        """
        if not PythonValidator.is_string(query):
            raise Exception('The provided "query" parameter is not a string.')

        return download_gif(query, Output.get_filename(output_filename, FileExtension.GIF))
    
    @staticmethod
    def download_google_drive_resource(
        google_drive_url: str,
        output_filename: Union[str, None] = None
    ):
        """
        Download the Google Drive resource from the given 'google_drive_url',
        if existing and available, to a local file called 'output_filename'.
        """
        if not PythonValidator.is_string(google_drive_url):
            raise Exception('The provided "google_drive_url" parameter is not a string.')
        
        # I don't know the exception of the file
        return GoogleDriveResource.download_from_url(google_drive_url, Output.get_filename(output_filename))
    
    @staticmethod
    def download_image(
        url: str,
        output_filename: Union[str, None] = None
    ) -> FileReturn:
        """
        Download the image from the provided 'url' and stores it, if
        existing and available, as a local file called 'output_filename'.
        """
        if not PythonValidator.is_string(url):
            raise Exception('The provided "url" parameter is not a string.')
        
        return download_image_2(url, Output.get_filename(output_filename))

    @staticmethod
    def download_video(
        url: str,
        output_filename: Union[str, None] = None
    ):
        """
        Download the video from the provided 'url' and stores it, if
        existing and available, as a local file called 'output_filename'.
        """
        if not PythonValidator.is_string(url):
            raise Exception('The provided "url" parameter is not a string.')
        
        return download_video(url, Output.get_filename(output_filename))
    
    @staticmethod
    def download_file(
        url: str,
        output_filename: str
    ):
        if not PythonValidator.is_string(url):
            raise Exception('The provided "url" parameter is not a string.')
        
        if output_filename is None:
            raise Exception(f'The "output_filename" parameter is mandatory as we do not know the file extension.')
        
        return download_file(url, output_filename)


from yta_general_utils.programming.decorators.singleton import _SingletonWrapper


class ClassPropertyDescriptor(object):
    """
    This class is based on this topic:
    - https://stackoverflow.com/a/5191224
    """

    def __init__(self, fget, fset = None):
        self.fget = fget
        self.fset = fset

    def __get__(self, obj, cls = None):
        if cls is None:
            cls = type(obj)
        return self.fget.__get__(obj, cls)()

    def __set__(self, obj, value):
        if not self.fset:
            raise AttributeError("can't set attribute")
        type_ = type(obj)
        return self.fset.__get__(obj, type_)(value)

    def setter(self, func):
        if not isinstance(func, (classmethod, staticmethod)):
            func = classmethod(func)
        self.fset = func
        return self

def classproperty(func):
    """
    Decorator to implement a class property.
    """
    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)

    return ClassPropertyDescriptor(func)


def singleton(
    cls: type
):
    """
    Singleton decorator that return a wrapper object
    that, when called, returns a single instance of
    the decorated class. For unit testing, use the
    '__wrapped__' attribute to access the class
    directly.
    """
    return _SingletonWrapper(cls)

def singleton_old(  
    cls
):
    """
    Decorator to implement a singleton class by
    making sure only one instance is returned each
    time the class is instantiated.

    You just need to use the decorator on top of
    the class you want to be singleton and you are
    ready to use it.

    How to declare and instantiate:

    @singleton
    class Singleton:
        pass

    s1 = Singleton()
    """
    instances = {}
    
    def get_instance(
        *args,
        **kwargs
    ):
        if cls not in instances:
            instances[cls] = cls(*args, **kwargs)

        return instances[cls]
    
    return get_instance


from yta_general_utils.programming.validator import PythonValidator
from yta_general_utils.programming.validator.error_message import ErrorMessage
from typing import Union
from random import choice as rand_choice
from enum import Enum


class YTAEnum(Enum):
    """
    Class to enhance the basic Enum functionality.
    """
    
    @classmethod
    def is_valid_value(
        cls,
        value: any
    ) -> bool:
        """
        This method returns True if the provided 'value' is a valid
        value for this YTAEnum class, or False if not.
        """
        return is_valid_value(value, cls)
    
    @classmethod
    def is_valid_name(
        cls,
        name: str,
        do_ignore_case: bool = False
    ) -> bool:
        """
        This method returns True if the provided 'name' is a valid
        name of this YTAEnum class, or False if not.
        """
        return is_valid_name(name, cls, do_ignore_case)
    
    @classmethod
    def is_valid_name_or_value(
        cls,
        name_or_value: any
    ) -> bool:
        """
        This method returns True if the provided 'name_or_value' is
        a valid name or a valid value of this YTAEnum class, or 
        False if not.
        """
        return is_valid_name_or_value(name_or_value, cls)
    
    @classmethod
    def is_valid(
        cls,
        name_or_value_or_enum: any
    ) -> bool:
        """
        This method returns True if the provided 'name_or_value_or_enum'
        is a valid name, a valid value or a valid instance of this
        YTAEnum class, or returns False if not.
        """
        return is_valid(name_or_value_or_enum, cls)
    
    @classmethod
    def name_to_enum(
        cls,
        name: str
    ) -> 'YTAEnum':
        """
        This method returns this 'enum' YTAEnum item instance if the
        provided 'name' is a valid name of that 'enum' YTAEnum class,
        or raises an Exception if not.
        """
        return from_name(name, cls)
    
    @classmethod
    def value_to_enum(
        cls,
        value: any
    ) -> 'YTAEnum':
        """
        This method returns the 'enum' YTAEnum item instance if the
        provided 'value' is a valid value of that 'enum' YTAEnum class,
        or raises an Exception if not.
        """
        return from_value(value, cls)
    
    @classmethod
    def name_or_value_to_enum(
        cls,
        name_or_value: any
    ) -> 'YTAEnum':
        """
        This method returns this 'enum' YTAEnum item instance if the
        provided 'name_or_value' is a valid name or a valid value of
        that 'enum' YTAEnum class, or raises an Exception if not.
        """
        return from_name_or_value(name_or_value, cls)
    
    @classmethod
    def to_enum(
        cls,
        name_or_value_or_enum: any
    ) -> 'YTAEnum':
        """
        This method returns this 'enum' YTAEnum item instance if the
        provided 'name_or_value_or_enum' is a valid name, a valid
        value or a valid instance of this 'enum' YTAEnum class, or
        raises an Exception if not.
        """
        return from_name_or_value_or_enum(name_or_value_or_enum, cls)

    @classmethod
    def get_all(
        cls
    ) -> list['YTAEnum']:
        """
        This method returns all the existing items in this 'cls' Enum
        class.
        """
        return get_all(cls)
    
    @classmethod
    def get_all_names(
        cls
    ) -> list[str]:
        """
        This method returns all the names of the existing items in
        this 'cls' Enum class.
        """
        return get_all_names(cls)

    @classmethod
    def get_all_values(
        cls
    ) -> list[any]:
        """
        This method returns all the values of the existing items in 
        this 'cls' Enum class.
        """
        return get_all_values(cls)
    
    @classmethod
    def get_all_values_as_str(
        cls
    ) -> str:
        """
        Get all the values of the existing items as strings separated
        by commas, useful to show as accepted values in errors.
        """
        return get_values_as_str(cls.get_all())
    
    @classmethod
    def get_all_names_as_str(
        cls
    ) -> str:
        """
        Get all the names of the existing items as strings separated
        by commas, useful to show as accepted values in errors.
        """
        return get_names_as_str(cls.get_all())
    
    @classmethod
    def get_valid_name(
        cls,
        name: str
    ) -> Union[str, None]:
        """
        This method will ignore cases to look for the provided 'name'
        as a valid name of the provided 'enums'. This method is useful
        when user provided us a name and we want to obtain the actual
        Enum name to be able to instantiate it, but maybe user provided
        'name' is quite different, invalid for instantiating but valid
        for our logic.

        This method returns None if the provided 'name' is not a valid
        name for this Enum class.
        """
        names = cls.get_all_names()

        try:
            return names[[
                enum_name.lower()
                for enum_name in names
            ].index(name.lower())]
        except Exception:
            return None
        
    @classmethod
    def get_random(
        cls
    ) -> 'YTAEnum':
        """
        Get an Enum item of the available ones chosen
        randomly.
        """
        return rand_choice(cls.get_all())
    
    @classmethod
    def default(
        cls
    ) -> 'YTAEnum':
        """
        Get the default value of the Enum class, that is, by
        default, the first one.

        You can overwrite this method to modify the default
        value.
        """
        return cls.get_all()[0]
    
    @classmethod
    def get_instance_if_name_is_accepted(
        cls,
        name: str,
        do_ignore_case: bool = False
    ) -> Union['YTAEnum', None]:
        """
        Get the Enum instance if the given 'name' is a
        valid name for this Enum instance.
        """
        names = (
            cls.get_all_names()
            if not do_ignore_case else
            [
                name.lower()
                for name in cls.get_all_names()
            ]
        )
        
        name = (
            name
            if not do_ignore_case else
            name.lower()
        )

        return (
            cls[cls.get_all_names()[names.index(name)]]
            if name in names else
            None
        )
    
    def get_value_if_accepted(
        self,
        value
    ) -> Union[any, None]:
        """
        Check if the provided 'value' is a value of the given Enum
        item. This will check the condition depending on the type
        of .value, that could be a single value, a list of values
        or a list of Enums.

        When the .value is a single value, it will check if that
        value is the given 'value'. When the .value is a list of
        values, it will check that the given 'value' is one of 
        those values. When the .value is a list of Enums, it will
        check that the provided 'value' is an instance of the 
        Enums in the list or the name or value of any of those.

        This method returns the single value, as it is, in this
        Enum value, so it will return an Enum if the this Enum
        value is a list of other Enums, or a single item.
        """
        # The name and the behaviour of this method is not the
        # most self-descriptive. This has been created to be
        # able to check if some value is a valid value of the
        # current Enum. For example, if working with an Enum
        # that holds other Enums of, for example, extensions,
        # we can check if the 'bmp' extension is accepted and
        # it will search through the different existing Enums
        # in .value and find it by the value Enum item .value
        # field, returning the whole Enum.
        if PythonValidator.is_list(self.value):
            if PythonValidator.is_enum_instance(self.value[0]):
                # It is a list of Enums, so use Enum values
                for enum in self.value:
                    # TODO: Maybe {enum.name, enum.value, enum} is faster
                    if value in [enum.name, enum.value, enum]:
                        return enum
            else:
                # It is a list of values, so we need to concat them
                if value in self.value:
                    return value
        else:
            if self.value == value:
                return value

        return None
    
    def is_accepted_value(
        self,
        value
    ) -> bool:
        """
        Check if the provided 'value' is a value of the given Enum
        item. This will check the condition depending on the type
        of .value, that could be a single value, a list of values
        or a list of Enums.

        When the .value is a single value, it will check if that
        value is the given 'value'. When the .value is a list of
        values, it will check that the given 'value' is one of 
        those values. When the .value is a list of Enums, it will
        check that the provided 'value' is an instance of the 
        Enums in the list or the name or value of any of those.

        This method returns True if the given 'value' is one of 
        the registered values, or False if not.
        """
        # The name and the behaviour of this method is not the
        # most self-descriptive. This has been created to be
        # able to check if some value is a valid value of the
        # current Enum. For example, if working with an Enum
        # that holds other Enums of, for example, extensions,
        # we can check if the 'bmp' extension is accepted and
        # it will search through the different existing Enums
        # in .value and find it by the value Enum item .value
        # field, returning True.
        return self.get_value_if_accepted(value) is not None
    
    @staticmethod
    def parse_as_enum(
        value,
        enum_classes: list['YTAEnum']
    ) -> 'YTAEnum':
        """
        Try to parse the provided 'value' as the also provided
        'enums' classes and return the corresponding Enum instance
        if found and valid.
        """
        if not PythonValidator.is_list(enum_classes):
            raise Exception('The provided "enums" parameter is not a list.')

        # TODO: This is not working well
        if any(not PythonValidator.is_enum_class(enum) for enum in enum_classes):
            raise Exception('Not all elements in the provided "enum_classes" list are YTAEnum classes.')
        
        for enum in enum_classes:
            try:
                return enum.to_enum(value)
            except:
                pass

        raise Exception(f'The provided "value" parameter "{str(value)}" is not parseable as one of the provided "enums".')




def is_enum(
    cls: Enum
) -> bool:
    """
    This method returns True if the provided 'cls' parameter is
    an enum class or subclass.
    """
    return (
        PythonValidator.is_instance(cls, Enum) or
        PythonValidator.is_subclass(cls, Enum) or
        PythonValidator.is_subclass(cls, YTAEnum)
    )

    # return (
    #     isinstance(cls, Enum) or
    #     issubclass(cls, (Enum, YTAEnum))
    # )

def is_valid(
    name_or_value_or_enum: any,
    enum: YTAEnum,
    do_ignore_case: bool = False
) -> bool:
    """
    Returns True if the provided 'name_or_value_or_enum' is
    a valid name or a valid value of the also provided 'enum'
    YTAEnum object, or even if it is an YTAEnum instance of
    that 'enum' YTAEnum class, or False if not.

    This method returns True or False.
    """
    return (
        PythonValidator.is_instance(name_or_value_or_enum, enum) or
        is_valid_name_or_value(name_or_value_or_enum, enum, do_ignore_case = do_ignore_case)
    )

def is_valid_name(
    name: any,
    enum: YTAEnum,
    do_ignore_case: bool = False
) -> bool:
    """
    Returns True if the provided 'name' is a valid name of
    the also provided 'enum' YTAEnum object, or False if not.

    This method returns True or False.
    """
    if not do_ignore_case:
        try:
            enum[name]

            return True
        except Exception:
            return False
    else:
        names = get_all_names()
        try:
            names[[
                enum_name.lower()
                for enum_name in names
            ].index(name.lower())]

            return True
        except Exception:
            return False
    
def is_valid_value(
    value: any,
    enum: YTAEnum
) -> bool:
    """
    Returns True if the provided 'value' is a valid value of
    the also provided 'enum' YTAEnum object.

    This method returns True or False.
    """
    try:
        if get_enum_from_value(value, enum):
            return True
    except Exception:
        return False
    
def is_valid_name_or_value(
    name_or_value: any,
    enum: YTAEnum,
    do_ignore_case: bool = False
) -> bool:
    """
    Returns True if the provided 'name_or_value' is a valid
    name or a valid value of the also provided 'enum' YTAEnum 
    object, or False if not.

    This method returns True or False.
    """
    return (
        is_valid_name(name_or_value, enum, do_ignore_case = do_ignore_case) or
        is_valid_value(name_or_value, enum)
    )
    
def from_name(
    name: any,
    enum: YTAEnum,
    do_ignore_case: bool = False
) -> 'YTAEnum':
    """
    This method returns the 'enum' YTAEnum item instance if the
    provided 'name' is a valid name of that 'enum' YTAEnum class,
    or raises an Exception if not.
    """
    if is_valid_name(name, enum, do_ignore_case):
        return enum[name]
    
    raise Exception(ErrorMessage.parameter_is_not_name_of_ytaenum_class(name, enum))

def from_value(
    value: any,
    enum: YTAEnum
) -> 'YTAEnum':
    """
    This method returns the 'enum' YTAEnum item instance if the
    provided 'value' is a valid value of that 'enum' YTAEnum class,
    or raises an Exception if not.
    """
    enum_from_value = get_enum_from_value(value, enum)
    if enum_from_value is not None:
        return enum_from_value
    
    raise Exception(ErrorMessage.parameter_is_not_value_of_ytaenum_class(value, enum))

def from_name_or_value(
    name_or_value: any,
    enum: YTAEnum,
    do_ignore_case: bool = False
) -> 'YTAEnum':
    """
    This method returns the 'enum' YTAEnum item instance if the
    provided 'name_or_value' is a valid name or a valid value of
    that 'enum' YTAEnum class, or raises an Exception if not.
    """
    try:
        enum_from_name = from_name(name_or_value, enum, do_ignore_case)

        return enum_from_name
    except:
        pass

    try:
        enum_from_value = from_value(name_or_value, enum)

        return enum_from_value
    except:
        pass

    raise Exception(ErrorMessage.parameter_is_not_name_nor_value_of_ytaenum_class(name_or_value, enum))

def from_name_or_value_or_enum(
    name_or_value_or_enum: any,
    enum: YTAEnum,
    do_ignore_case: bool = False
) -> 'YTAEnum':
    """
    This method returns the 'enum' YTAEnum item instance if the
    provided 'name_or_value_or_enum' is a valid name, a valid
    value or a valid instance of that 'enum' YTAEnum class, or
    raises an Exception if not.
    """
    if PythonValidator.is_instance(name_or_value_or_enum, enum):
        return name_or_value_or_enum
    else:
        try:
            enum_from_name_or_value = from_name_or_value(name_or_value_or_enum, enum, do_ignore_case)

            return enum_from_name_or_value
        except:
            pass
    
    raise Exception(ErrorMessage.parameter_is_not_name_nor_value_nor_enum_of_ytaenum_class(name_or_value_or_enum, enum))
    
def get_enum_from_value(
    value: any,
    enum: YTAEnum
) -> 'YTAEnum':
    """
    Check if the provided 'value' is a valid value of the given
    'enum'. This method will check if the 'enum.value' is a
    simple value, a list of values or even a list of enums and
    will make the corresponding verifications.

    This method returns the whole Enum instance or None if not
    a valid value.
    """
    try:
        first = enum.get_all()[0]
        enum_type = 'normal'
        if PythonValidator.is_list(first.value):
            enum_type = 'list'
            if PythonValidator.is_instance(first.value[0], Enum):
                enum_type = 'list_of_enums'

        if enum_type == 'normal':
            return enum(value)
        else:
            for value_enum in enum.get_all():
                if enum_type == 'list_of_enums':
                    # We detect if the provided 'value' is .value, .name or even
                    # an instance of this Enum
                    if any(value in [enum.name, enum.value, enum] for enum in value_enum.value):
                        return enum(value_enum.value)
                elif enum_type == 'list':
                    # We detect if the provided 'value' is any of the elements
                    # in the .value list
                    if value in value_enum.value:
                        return enum(value_enum.value)
    except:
        pass
                
    return None

def is_name_or_value(
    name_or_value: any,
    enum: YTAEnum
) -> bool:
    """
    This method validates if the provided 'value' is a name
    or a value of the also provided 'enum' YTAEnum, raising
    an Exception if not.

    This method returns the enum item (containing .name and
    .value) if it is valid.
    """
    if name_or_value in enum.get_all_names():
        return enum[name_or_value]
    elif name_or_value in enum.get_all_values():
        # TODO: This won't work for list or Enum list
        return enum(name_or_value)
    
    raise Exception(ErrorMessage.parameter_is_not_name_nor_value_of_ytaenum_class(name_or_value, enum))
    
def get_all(
    enum: Enum
) -> list['YTAEnum']:
    """
    This method returns all the items defined in a Enum subtype that
    is provided as the 'enum' parameter.
    """
    if not enum:
        raise Exception('No "enum" provided.')
    
    if (
        not PythonValidator.is_instance(enum, Enum) and
        not PythonValidator.is_subclass(enum, Enum)
    ):
        raise Exception('The "enum" parameter provided is not an Enum.')
    
    return [
        item
        for item in enum
    ]

def get_all_names(
    cls
) -> list[str]:
    """
    Returns a list containing all the registered enum names.
    """
    return [
        item.name
        for item in get_all(cls)
    ]

def get_all_names_as_str(
    cls
) -> str:
    """
    Returns a string containing all the enums of the provided
    YTAEnum 'cls' class names separated by commas.
    """
    return get_names_as_str(get_all(cls))

def get_all_values(
    cls
) -> list[any]:
    """
    Returns a list containing all the registered enum values.
    """
    return [
        item.value
        for item in get_all(cls)
    ]

def get_all_values_as_str(
    cls
) -> str:
    """
    Returns a string containing all the enums of the provided
    YTAEnum 'cls' class values separated by commas.
    """
    # TODO: The stringify process has to change
    return get_values_as_str(get_all(cls))

def get_names(
    enums: list[Enum]
) -> list[str]:
    """
    Returns a list containing all the names of the provided
    'enums' Enum elements.
    """
    if any(not is_enum(enum) for enum in enums):
        raise TypeError('At least one of the given "enums" is not an Enum class or subclass.')    
    
    return [
        item.name
        for item in enums
    ]

def get_names_as_str(
    enums: list[YTAEnum]
) -> str:
    """
    Returns a string containing the provided 'enums' names separated
    by commas.
    """
    if any(not is_enum(enum) for enum in enums):
        raise TypeError('At least one of the given "enums" is not an Enum class or subclass.')    
    
    return ', '.join(get_names(enums))

def get_values(
    enums: list[Enum]
) -> list[any]:
    """
    Returns a list containing all the values of the provided
    'enums' Enum elements.
    """
    if any(not is_enum(enum) for enum in enums):
        raise TypeError('At least one of the given "enums" is not an Enum class or subclass.')    
    
    values = []
    for enum in enums:
        if PythonValidator.is_list(enum.value):
            if PythonValidator.is_instance(enum.value[0], Enum):
                # It is a list of Enums, so use Enum values
                values.append(', '.join([enum_value.value for enum_value in enum.value]))
                # TODO: Maybe the whole Enum class+name (?)
                #values.append(', '.join([f'{enum.__class__}.{enum_value.name}' for enum_value in enum.value]))
            else:
                # It is a list of values, so we need to concat them
                values.append(', '.join(enum.value))
        else:
            # It is just one value
            values.append(enum.value)

    return values

def get_values_as_str(
    enums: Enum
) -> str:
    """
    Returns a string containing the provided 'enums' values separated
    by commas.
    """
    if any(not is_enum(enum) for enum in enums):
        raise TypeError('At least one of the given "enums" is not an Enum class or subclass.')    
     
    # TODO: The stringify method has to change
    return ', '.join(get_values(enums))


from yta_general_utils.random import Random
from time import sleep

import pyautogui


def long_wait():
    """
    Waits from 2 to 6 seconds, moving the cursor accidentally 3 times.
    """
    wait(Random.float_between(2, 6, 0.001))
    # from random import randint
    # wait(randint(2000, 6000) * 0.001)

def short_wait():
    """
    Waits from 0.4 to 2 seconds, moving the cursor accidentally 3 times.
    """
    wait(Random.float_between(0.4, 2, 0.001))
    # from random import randint
    # wait(randint(40, 200) * 0.01)

def wait(time):
    """
    Waits for the provided 'time' in seconds and
    accidentally moves the cursor 3 times.
    """
    sleep_factors = [0.23, 0.48, 0.29]
    for i in range(3):
        sleep(time * sleep_factors[i])
        x_movement = Random.int_between(1, 7)
        if Random.bool():
            x_movement = -x_movement
        y_movement = Random.int_between(1, 7)
        if Random.bool():
            y_movement = -y_movement
        pyautogui.move(x_movement, y_movement)

def write(text):
    for char in text:
        pyautogui.write(char)
        sleep(Random.float_between(0.05, 0.2, 0.001))

def scroll(height):
    times = height % 200
    for i in range(times):
        pyautogui.scroll(height / times)
        sleep(Random.float_between(0.01, 0.08, 0.001))

def click_element(driver, element):
    # TODO: Avoid passing the driver please, or refactor if possible
    move_to_element(driver, element)
    position = pyautogui.position()
    pyautogui.click(position[0], position[1])

def move_to_element(chromedriver, element):
    panel_height = chromedriver.execute_script('return window.outerHeight - window.innerHeight;')
    panel_width = chromedriver.execute_script('return window.outerWidth - window.innerWidth;')
    print('panel height: ' + str(panel_height))
    print('panel width: ' + str(panel_width))
    print('y: ' + str(element.location['y']))
    print('x: ' + str(element.location['x']))
    print('width: ' + str(element.size['width']))
    print('height: ' + str(element.size['height']))
    x = element.location['x'] + panel_width + (element.size['width'] / 2)
    y = element.location['y'] + panel_height + (element.size['height'] / 2)
    print('x: ' + str(x))
    print('y: ' + str(y))
    move_cursor_to(x, y)

def move_cursor_to(x, y):
    # We can fail and then get back
    x1, y1 = pyautogui.position()  # Starting position
    # (0, 0) is upper left
    x_error = Random.int_between(0, int(0.2 * abs(x - x1)))
    if Random.bool():
        x_error = -x_error
    y_error = Random.int_between(0, int(0.2 * abs(y - y1)))
    if Random.bool():
        y_error = -y_error

    # We move first to the position with an error
    __move_cursor_to(x + x_error, y + y_error)
    # We spend some time in realizing we failed (60 to 230ms)
    sleep(Random.float_between(0.06, 0.23, 0.001))
    # We then move to the expected position
    __move_cursor_to(x, y)

def __move_cursor_to(x2, y2):
    import numpy as np
    import time
    from scipy import interpolate
    import math

    # Any duration less than this is rounded to 0.0 to instantly move the mouse.
    pyautogui.MINIMUM_DURATION = 0  # Default: 0.1
    # Minimal number of seconds to sleep between mouse moves.
    pyautogui.MINIMUM_SLEEP = 0  # Default: 0.05
    # The number of seconds to pause after EVERY public function call.
    pyautogui.PAUSE = 0  # Default: 0.1

    def point_dist(x1,y1,x2,y2):
        return math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)
    
    cp = Random.int_between(2, 5)  # Number of control points. Must be at least 2.
    x1, y1 = pyautogui.position()  # Starting position

    # Distribute control points between start and destination evenly.
    x = np.linspace(x1, x2, num = cp, dtype = 'int')
    y = np.linspace(y1, y2, num = cp, dtype = 'int')

    # Randomise inner points a bit (+-RND at most).
    RND = 10
    xr = [
        Random.int_between(-RND, RND)
        for _ in range(cp)
    ]
    yr = [
        Random.int_between(-RND, RND)
        for _ in range(cp)
    ]
    xr[0] = yr[0] = xr[-1] = yr[-1] = 0
    x += xr
    y += yr

    # Approximate using Bezier spline.
    degree = 3 if cp > 3 else cp - 1  # Degree of b-spline. 3 is recommended.
    # Must be less than number of control points.
    tck, u = interpolate.splprep([x, y], k = degree)
    # Move upto a certain number of points
    u = np.linspace(0, 1, num = 2 + int(point_dist(x1, y1, x2, y2) / 50.0))
    points = interpolate.splev(u, tck)

    # Move mouse.
    duration = 0.1
    timeout = duration / len(points[0])
    point_list=zip(*(i.astype(int) for i in points))
    for point in point_list:
        #pyautogui.moveTo(*point, uniform(0.6, 1.7), pyautogui.easeOutQuad)
        pyautogui.moveTo(*point)
        time.sleep(timeout + Random.int_between(0, 1) * 0.005)


from yta_general_utils.web.scraper.chrome_scraper import ChromeScraper
from yta_general_utils.programming.validator.parameter import ParameterValidator
from typing import Union

import time


class WebRealTimeAudioTranscriptor:
    """
    Class to wrap a functionality related to real
    time audio transcription by using a web scraper.

    This class uses local files to create a simple
    web page that uses the chrome speech recognition
    to get the transcription.
    """

    _URL: str = 'file:///C:/Users/dania/Downloads/JS-TRANSCRIBE/transcribe.html'
    """
    The url to our local web page file.
    """
    max_waiting_time: Union[float, None]
    """
    The maximum time the software will be waiting
    to detect an audio transcription before exiting
    with an empty result.
    """

    def __init__(
        self,
        max_waiting_time: Union[float, None] = 15.0
    ):
        ParameterValidator.validate_positive_float('max_waiting_time', max_waiting_time, do_include_zero = True)

        self.scraper = ChromeScraper()
        self.max_waiting_time = (
            9999 # TODO: This is risky if no microphone or something
            if (
                max_waiting_time == 0 or
                max_waiting_time is None
            ) else
            max_waiting_time
        )

    def _load(
        self
    ):
        """
        Navigates to the web page when not yet on it.
        """
        if self.scraper.current_url != self._URL:
            self.scraper.go_to_web_and_wait_until_loaded(self._URL)

    def _get_transcription(
        self
    ):
        """
        Get the text that has been transcripted from the
        audio.
        """
        self._load()

        return self.scraper.find_element_by_id('result').get_attribute('value')
    
    def _click_transcription_button(
        self
    ):
        """
        Performa click on the button that enables (or
        disables) the microphone so it starts (or ends)
        transcribing the text.
        """
        self._load()

        self.scraper.find_element_by_id('toggle').click()
    
    def transcribe(
        self
    ) -> str:
        """
        A web scraper instance loads the internal web
        that uses the Chrome speech recognition to get
        the audio transcription, by pressing the
        button, waiting for audio input through the
        microphone, and pressing the button again.

        If the page was previously loaded it won't be
        loaded again.
        """
        self._load()

        WAITING_TIME = 0.25

        self._click_transcription_button()

        time_elapsed = 0
        while time_elapsed < self.max_waiting_time:
            transcription = self._get_transcription()
            if transcription != '':
                # We have it, force exit the loop
                time_elapsed = self.max_waiting_time
            else:
                time.sleep(WAITING_TIME)
                time_elapsed += WAITING_TIME

        self._click_transcription_button()

        return transcription



